﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using NPersistence;
using NHibernate;
using NPersistence.NHibernate.transaction;
using NPersistence.NHibernate.query;
using NPersistence.NHibernate.util;
using log4net;
using NPersistence.spi;
using NHibernate.Transaction;
using NHibernate.Engine;
using NPersistence.NHibernate;

namespace NPersistence.NHibernate
{
    class EntityManagerImpl : EntityManager
    {
        private ILog logger = LogManager.GetLogger("ObjectDB.Npa");

        private static List<String> m_entityManagerSpecificProperties = new List<String>()
        {
            AvailableSettings.LOCK_SCOPE, 
            AvailableSettings.LOCK_SCOPE_JPA, 
            AvailableSettings.LOCK_TIMEOUT, 
            AvailableSettings.LOCK_SCOPE_JPA,
            AvailableSettings.FLUSH_MODE, 
            AvailableSettings.FLUSH_MODE_JPA, 
            AvailableSettings.SHARED_CACHE_RETRIEVE_MODE_JPA,
            AvailableSettings.SHARED_CACHE_RETRIEVE_MODE,
            AvailableSettings.SHARED_CACHE_STORE_MODE_JPA,
            AvailableSettings.SHARED_CACHE_STORE_MODE,
            QueryHints.SPEC_HINT_TIMEOUT,
            QueryHints.SPEC_HINT_TIMEOUT2
        };

	   
        private ISession m_session;
        private Dictionary<string, object> m_properties;
        private EntityManagerFactory m_entityManagerFactory;
        private ISessionFactory m_sessionFactory;
        
        [NonSerialized]
        protected  EntityTransactionImpl m_tx;
        //JTA is not supported yet
        //private  PersistenceUnitTransactionType m_transactionType; // JTA or regular
        private bool m_open;
        private Type m_sessionInterceptorClass;
        private PersistenceContextType m_persistenceContextType;
        private bool m_discardOnClose;

        public EntityManagerImpl(ISessionFactory sessionFactory, 
                                 Dictionary<object, object> properties, 
                                 EntityManagerFactory entityManagerFactory, 
                                 PersistenceUnitTransactionType transactionType,
                                 PersistenceContextType persistenceContextType,
                                 Type sessionInterceptorClass,
                                 bool discardOnClose)
        {
            m_open = true;
            m_sessionFactory = sessionFactory;
            m_entityManagerFactory = entityManagerFactory;
            //m_transactionType =  transactionType;
            m_persistenceContextType = persistenceContextType;
            m_sessionInterceptorClass = sessionInterceptorClass;
            m_discardOnClose = discardOnClose;
            m_tx = new EntityTransactionImpl(this);
            m_properties = new Dictionary<string, object>();
            if (properties != null)
            {
                foreach (string key in m_entityManagerSpecificProperties)
                {
                    if (properties.ContainsKey(key))
                    {
                        m_properties.Add(key, properties[key]);
                    }
                }
            }
        }

        public ISession getSession()
        {
            if (!m_open)
            {
                throw new InvalidOperationException("EntityManager is closed");
            }
            return getRawSession();
        }

        protected ISession getRawSession()
        {
            if (m_session == null)
            {
                if (m_sessionInterceptorClass != null)
                {
                    IInterceptor interceptor = null;
                    try
                    {
                        interceptor = (IInterceptor)m_sessionInterceptorClass.GetConstructor(new Type[] { }).Invoke(new object[] { });
                    }
                    catch (InvalidCastException e)
                    {
                        throw new PersistenceException(
                                "Session interceptor does not implement Interceptor: "
                                        + m_sessionInterceptorClass, e);
                    }
                    catch (Exception e)
                    {
                        throw new PersistenceException(
                                "Unable to instanciate session interceptor: "
                                        + m_sessionInterceptorClass, e);
                    }
                    m_session = m_sessionFactory.OpenSession(interceptor);
                }
                else
                {
                    m_session = m_sessionFactory.OpenSession();
                }
               
            }
            return m_session;
        }

        private void checkTransactionNeeded()
        {
            if (m_persistenceContextType == PersistenceContextType.TRANSACTION && !isTransactionInProgress())
            {
                //no need to mark as rollback, no tx in progress
                throw new TransactionRequiredException(
                        "no transaction is in progress for a TRANSACTION type persistence context"
                );
            }
        }

        public bool isTransactionInProgress()
        {
            return getRawSession().Transaction.IsActive;
        }

        public void Persist(object entity)
        {
            try
            {
                getSession().Save(entity);
            }
            catch (MappingException e)
            {
                throw new ArgumentException(e.Message);
            }
            catch (Exception e)
            {
                throw convert(e);
            }
        }

        public Exception convert(Exception e) 
        {
		    Exception result = e;
		    if ( e is HibernateException ) 
            {
			    result = convert( ( HibernateException ) e );
		    }
		    else {
			    markAsRollback();
		    }
		    return result;
	    }

        public Exception convert(HibernateException e) 
        {
		    if ( e is StaleStateException ) {
			    PersistenceException converted = wrapStaleStateException( ( StaleStateException ) e );
			    handlePersistenceException( converted );
			    return converted;
		    }
		    else if ( e is ObjectNotFoundException ) {
			    EntityNotFoundException converted = new EntityNotFoundException( e.Message );
			    handlePersistenceException( converted );
			    return converted;
		    }
		    else if ( e is global::NHibernate.NonUniqueResultException ) 
            {
                NPersistence.NonUniqueResultException converted = new NPersistence.NonUniqueResultException(e.Message);
			    handlePersistenceException( converted );
			    return converted;
		    }
		    else if ( e is UnresolvableObjectException ) 
            {
			    EntityNotFoundException converted = new EntityNotFoundException( e.Message );
			    handlePersistenceException( converted );
			    return converted;
		    }
		    else if ( e is QueryException ) 
            {
			    return new ArgumentException(e.Message, e );
		    }
		    else if ( e is TransientObjectException ) 
            {
			    try 
                {
				    markAsRollback();
			    }
			    catch ( Exception ne ) {
				    //we do not want the subsequent exception to swallow the original one
				    logger.Error( "Unable to mark for rollback on TransientObjectException: ", ne );
			    }
                return new InvalidOperationException(e.Message, e); //Spec 3.2.3 Synchronization rules
		    }
		    else 
            {
			    PersistenceException converted = new PersistenceException( e.Message, e );
			    handlePersistenceException( converted );
			    return converted;
		    }
	    }

        public void handlePersistenceException(PersistenceException e) 
        {
		    if ( e is NoResultException ) 
            {
			    return;
		    }
		    if ( e is NPersistence.NonUniqueResultException ) 
            {
			    return;
		    }
		    if ( e is LockTimeoutException ) 
            {
			    return;
		    }
		    if ( e is QueryTimeoutException ) 
            {
			    return;
		    }

		    try 
            {
			    markAsRollback();
		    }
		    catch ( Exception ne ) 
            {
			    //we do not want the subsequent exception to swallow the original one
			    logger.Error( "Unable to mark for rollback on PersistenceException: ", ne );
		    }
	    }

        public T Merge<T>(T entity) where T : class
        {
            checkTransactionNeeded();
            try
            {
                return (T)m_session.Merge(entity);
            }
            catch (ObjectDeletedException sse)
            {
                throw new ArgumentException(sse.Message, sse);
            }
            catch (MappingException e)
            {
                throw new ArgumentException(e.Message, e);
            }
            catch (Exception e)
            { //including HibernateException
                throw convert(e);
            }
        }


        public void Remove(object entity)
        {
            checkTransactionNeeded();
            try
            {
                getSession().Delete(entity);
            }
            catch (MappingException e)
            {
                throw new ArgumentException(e.Message, e);
            }
            catch (Exception e)
            { //including HibernateException
                throw convert(e);
            }
        }

        public void throwPersistenceException(PersistenceException e) 
        {
		    if ( ! ( e is NoResultException ) && !( e is NPersistence.NonUniqueResultException ) ) 
            {
			    try {
				    markAsRollback();
			    }
			    catch (Exception ne) {
				    //we do not want the subsequent exception to swallow the original one
				    logger.Error( "Unable to mark for rollback on PersistenceException: ", ne);
			    }
		    }
		    throw e;
	    }

        public void throwPersistenceException(HibernateException e) 
        {
		    if ( e is StaleStateException ) 
            {
			    PersistenceException pe = wrapStaleStateException( (StaleStateException) e );
			    throwPersistenceException( pe );
		    }
		    else if ( e is ObjectNotFoundException ) 
            {
			    throwPersistenceException( new EntityNotFoundException( e.Message) );
		    }
		    else if ( e is global::NHibernate.NonUniqueResultException ) 
            {
			    throwPersistenceException( new NPersistence.NonUniqueResultException( e.Message ) );
		    }
		    else if ( e is UnresolvableObjectException ) 
            {
			    throwPersistenceException( new EntityNotFoundException( e.Message ) );
		    }
		    else if ( e is QueryException ) 
            {
			    throw new ArgumentException(e.Message, e );
		    }
		    else if ( e is TransientObjectException ) 
            {
			    try 
                {
				    markAsRollback();
			    }
			    catch (Exception ne) 
                {
				    //we do not want the subsequent exception to swallow the original one
				    logger.Error( "Unable to mark for rollback on TransientObjectException: ", ne);
			    }
                throw new InvalidOperationException(e.Message); //Spec 3.2.3 Synchronization rules
		    }
		    else 
            {
			    throwPersistenceException( new PersistenceException( e ) );
		    }
	    }

        public PersistenceException wrapStaleStateException(StaleStateException e) 
        {
		    PersistenceException pe;
		    if ( e is StaleObjectStateException ) {
			    StaleObjectStateException sose = (StaleObjectStateException) e;
                object identifier = sose.Identifier;
			    if (identifier != null) 
                {
				    Object entity = getRawSession().Load( sose.EntityName, identifier );
				    if ( entity != null) 
                    {
					    //avoid some user errors regarding boundary crossing
					    pe = new OptimisticLockException( e.Message, e, entity );
				    }
				    else 
                    {
					    pe = new OptimisticLockException( e.Message, e );
				    }
			    }
			    else 
                {
				    pe = new OptimisticLockException( e.Message, e );
			    }
		    }
		    else 
            {
			    pe = new OptimisticLockException( e.Message, e);
		    }
		    return pe;
	    }

        protected void markAsRollback()
        {
            logger.Debug("mark transaction for rollback");
            if (m_tx.IsActive())
            {
                m_tx.SetRollbackOnly();
            }
            // JTA is not supported yet.
            //else
            //{
            //    //no explicit use of the tx. boundaries methods
            //    if (PersistenceUnitTransactionType.JTA == m_transactionType)
            //    {

            //        ITransactionFactory transactionManager =
            //                ((ISessionFactoryImplementor)getRawSession().SessionFactory).TransactionFactory;
            //        if (transactionManager == null)
            //        {
            //            throw new PersistenceException(
            //                    "Using a JTA persistence context wo setting hibernate.transaction.manager_lookup_class"
            //            );
            //        }
            //        try
            //        {
            //            transactionManager. setRollbackOnly();
            //        }
            //        catch (SystemException e)
            //        {
            //            throw new PersistenceException("Unable to set the JTA transaction as RollbackOnly", e);
            //        }
            //    }
            //}
        }

        /**
         * Find by primary key.
         * Search for an entity of the specified class and primary key.
         * If the entity instance is contained in the persistence context,
         * it is returned from there.
         * @param entityClass  entity class
         * @param primaryKey  primary key
         * @return the found entity instance or null if the entity does
         *         not exist
         * @throws ArgumentException if the first argument does
         *         not denote an entity type or the second argument is 
         *         is not a valid type for that entity’s primary key or
         *         is null
         */
        public T Find<T>(System.Type entityClass, object primaryKey)
        {
            return innerFind<T>(entityClass, primaryKey, null, null);
        }

        /**
         * Find by primary key, using the specified properties. 
         * Search for an entity of the specified class and primary key. 
         * If the entity instance is contained in the persistence 
         * context, it is returned from there. 
         * If a vendor-specific property or hint is not recognized, 
         * it is silently ignored. 
         * @param entityClass  entity class
         * @param primaryKey   primary key
         * @param properties  standard and vendor-specific properties 
         *        and hints
         * @return the found entity instance or null if the entity does
         *         not exist 
         * @throws ArgumentException if the first argument does 
         *         not denote an entity type or the second argument is
         *         is not a valid type for that entity’s primary key or 
         *         is null 
         * @since Java Persistence 2.0
         */
        public T Find<T>(System.Type entityClass, object primaryKey,
                          Dictionary<string, object> properties)
        {
            return innerFind<T>(entityClass, primaryKey, null, properties);
        }

        /**
         * Find by primary key and lock.
         * Search for an entity of the specified class and primary key
         * and lock it with respect to the specified lock type.
         * If the entity instance is contained in the persistence context,
         * it is returned from there, and the effect of this method is
         * the same as if the lock method had been called on the entity.
         * <p> If the entity is found within the persistence context and the
         * lock mode type is pessimistic and the entity has a version
         * attribute, the persistence provider must perform optimistic
         * version checks when obtaining the database lock.  If these 
         * checks fail, the <code>OptimisticLockException</code> will be thrown.
         * <p>If the lock mode type is pessimistic and the entity instance
         * is found but cannot be locked:
         * <ul>
         * <li> the <code>PessimisticLockException</code> will be thrown if the database
         *    locking failure causes transaction-level rollback
         * <li> the <code>LockTimeoutException</code> will be thrown if the database
         *    locking failure causes only statement-level rollback
         * </ul>
         * @param entityClass  entity class
         * @param primaryKey  primary key
         * @param lockMode  lock mode
         * @return the found entity instance or null if the entity does
         *         not exist
         * @throws ArgumentException if the first argument does
         *         not denote an entity type or the second argument is 
         *         not a valid type for that entity's primary key or 
         *         is null
         * @throws TransactionRequiredException if there is no 
         *         transaction and a lock mode other than NONE is
         *         specified
         * @throws OptimisticLockException if the optimistic version 
         *         check fails
         * @throws PessimisticLockException if pessimistic locking 
         *         fails and the transaction is rolled back
         * @throws LockTimeoutException if pessimistic locking fails and
         *         only the statement is rolled back
         * @throws PersistenceException if an unsupported lock call 
         *         is made
         * @since Java Persistence 2.0
         */
        public T Find<T>(System.Type entityClass, object primaryKey,
                          LockModeType lockMode)
        {
            return innerFind<T>(entityClass, primaryKey, lockMode, null);
        }

        /**
         * Find by primary key and lock, using the specified properties. 
         * Search for an entity of the specified class and primary key
         * and lock it with respect to the specified lock type.
         * If the entity instance is contained in the persistence context,
         * it is returned from there.  
         * <p> If the entity is found
         * within the persistence context and the lock mode type
         * is pessimistic and the entity has a version attribute, the
         * persistence provider must perform optimistic version checks
         * when obtaining the database lock.  If these checks fail,
         * the <code>OptimisticLockException</code> will be thrown.
         * <p>If the lock mode type is pessimistic and the entity instance
         * is found but cannot be locked:
         * <ul>
         * <li> the <code>PessimisticLockException</code> will be thrown if the database
         *    locking failure causes transaction-level rollback
         * <li> the <code>LockTimeoutException</code> will be thrown if the database
         *    locking failure causes only statement-level rollback
         * </ul>
         * <p>If a vendor-specific property or hint is not recognized, 
         * it is silently ignored.  
         * <p>Portable applications should not rely on the standard timeout
         * hint. Depending on the database in use and the locking
         * mechanisms used by the provider, the hint may or may not
         * be observed.
         * @param entityClass  entity class
         * @param primaryKey  primary key
         * @param lockMode  lock mode
         * @param properties  standard and vendor-specific properties
         *        and hints
         * @return the found entity instance or null if the entity does
         *         not exist
         * @throws ArgumentException if the first argument does
         *         not denote an entity type or the second argument is 
         *         not a valid type for that entity's primary key or 
         *         is null
         * @throws TransactionRequiredException if there is no 
         *         transaction and a lock mode other than <code>NONE</code> is 
         *         specified
         * @throws OptimisticLockException if the optimistic version 
         *         check fails
         * @throws PessimisticLockException if pessimistic locking 
         *         fails and the transaction is rolled back
         * @throws LockTimeoutException if pessimistic locking fails and
         *         only the statement is rolled back
         * @throws PersistenceException if an unsupported lock call 
         *         is made
         * @since Java Persistence 2.0
         */
        public T Find<T>(System.Type entityClass, object primaryKey,
                          LockModeType lockMode,
                          Dictionary<string, object> properties)
        {
            return innerFind<T>(entityClass, primaryKey, lockMode, properties);
        }

        private T innerFind<T>(System.Type entityClass, object primaryKey,
                          LockModeType? lockMode,
                          Dictionary<string, object> properties)
        {
            CacheMode previousCacheMode = getSession().CacheMode;
            CacheMode cacheMode = determineAppropriateLocalCacheMode(properties);
            try
            {
                getSession().CacheMode = cacheMode;
                if (lockMode.HasValue)
                {
                    LockMode lm = LockModeTypeHelper.getLockMode(lockMode.Value);
                    return (T)getSession().Get(entityClass, primaryKey, lm);
                }
                else
                {
                    return (T)getSession().Get(entityClass, primaryKey);
                }
            }
            catch (ObjectDeletedException e)
            {
                //the spec is silent about people doing remove() find() on the same PC
                return default(T);
            }
            catch (ObjectNotFoundException e)
            {
                //should not happen on the entity itself with get
                throw new ArgumentException(e.Message, e);
            }
            catch (MappingException e)
            {
                throw new ArgumentException(e.Message, e);
            }
            catch (TypeMismatchException e)
            {
                throw new ArgumentException(e.Message, e);
            }
            catch (InvalidCastException e)
            {
                throw new ArgumentException(e.Message, e);
            }
            catch (HibernateException he)
            {
                throw convert(he);
            }
            finally
            {
                getSession().CacheMode = previousCacheMode;
            }
        }

        public CacheMode determineAppropriateLocalCacheMode(Dictionary<String, Object> localProperties)
        {
            CacheRetrieveMode? retrieveMode = null;
            CacheStoreMode? storeMode = null;
            if (localProperties != null)
            {
                retrieveMode = determineCacheRetrieveMode(localProperties);
                storeMode = determineCacheStoreMode(localProperties);
            }
            if (retrieveMode == null)
            {
                // use the EM setting
                retrieveMode = determineCacheRetrieveMode(this.m_properties);
            }
            if (storeMode == null)
            {
                // use the EM setting
                storeMode = determineCacheStoreMode(this.m_properties);
            }
            return CacheModeHelper.interpretCacheMode(storeMode, retrieveMode);
        }

        private CacheRetrieveMode? determineCacheRetrieveMode(Dictionary<String, Object> settings)
        {
            if (settings.ContainsKey(AvailableSettings.SHARED_CACHE_RETRIEVE_MODE_JPA))
            {
                return (CacheRetrieveMode)settings[AvailableSettings.SHARED_CACHE_RETRIEVE_MODE_JPA];
            }
            if (settings.ContainsKey(AvailableSettings.SHARED_CACHE_RETRIEVE_MODE))
            {
                return (CacheRetrieveMode)settings[AvailableSettings.SHARED_CACHE_RETRIEVE_MODE];
            }
            return null;
        }

        private CacheStoreMode? determineCacheStoreMode(Dictionary<String, Object> settings)
        {
            if (settings.ContainsKey(AvailableSettings.SHARED_CACHE_STORE_MODE_JPA))
            {
                return (CacheStoreMode)settings[AvailableSettings.SHARED_CACHE_STORE_MODE_JPA];
            }
            if (settings.ContainsKey(AvailableSettings.SHARED_CACHE_STORE_MODE))
            {
                return (CacheStoreMode)settings[AvailableSettings.SHARED_CACHE_STORE_MODE];
            }
            return null;
        }

        /**
         * Get an instance, whose state may be lazily fetched.
         * If the requested instance does not exist in the database,
         * the <code>EntityNotFoundException</code> is thrown when the instance 
         * state is first accessed. (The persistence provider runtime is 
         * permitted to throw the <code>EntityNotFoundException</code> when 
         * <code>getReference</code> is called.)
         * The application should not expect that the instance state will
         * be available upon detachment, unless it was accessed by the
         * application while the entity manager was open.
         * @param entityClass  entity class
         * @param primaryKey  primary key
         * @return the found entity instance
         * @throws ArgumentException if the first argument does
         *         not denote an entity type or the second argument is
         *         not a valid type for that entity’s primary key or
         *         is null
         * @throws EntityNotFoundException if the entity state 
         *         cannot be accessed
         */
        public T GetReference<T>(System.Type entityClass,
                                      object primaryKey)
        {
            try
            {
                return (T)getSession().Load(entityClass, primaryKey);
            }
            catch (MappingException e)
            {
                throw new ArgumentException(e.Message, e);
            }
            catch (TypeMismatchException e)
            {
                throw new ArgumentException(e.Message, e);
            }
            catch (InvalidCastException e)
            {
                throw new ArgumentException(e.Message, e);
            }
            catch (HibernateException he)
            {
                throw convert(he);
            }
        }

        /**
         * Synchronize the persistence context to the
         * underlying database.
         * @throws TransactionRequiredException if there is
         *         no transaction
         * @throws PersistenceException if the flush fails
         */
        public void Flush()
        {
            try
            {
                if (!isTransactionInProgress())
                {
                    throw new TransactionRequiredException("no transaction is in progress");
                }
                getSession().Flush();
            }
            catch (Exception e)
            {
                throw convert(e);
            }
        }

        /**
         * Set the flush mode that applies to all objects contained
         * in the persistence context. 
         * @param flushMode  flush mode
         */
        public void SetFlushMode(FlushModeType flushMode)
        {
            if (flushMode == FlushModeType.AUTO)
            {
                m_session.FlushMode = FlushMode.Auto;
            }
            else if (flushMode == FlushModeType.COMMIT)
            {
                m_session.FlushMode = FlushMode.Commit;
            }
            else
            {
                throw new ArgumentException("Unexpected argument: " + flushMode);
            }
        }

        /**
         * Get the flush mode that applies to all objects contained
         * in the persistence context. 
         * @return flushMode
         */
        public FlushModeType GetFlushMode()
        {
            if (m_session.FlushMode == FlushMode.Auto)
            {
                return FlushModeType.AUTO;
            }
            else if (m_session.FlushMode == FlushMode.Commit)
            {
                return FlushModeType.COMMIT;
            }
            else
            {
                throw new InternalException("How come flush mode on the session is: " + m_session.FlushMode);
            }
        }

        /**
         * Lock an entity instance that is contained in the persistence
         * context with the specified lock mode type.
         * <p>If a pessimistic lock mode type is specified and the entity
         * contains a version attribute, the persistence provider must 
         * also perform optimistic version checks when obtaining the 
         * database lock.  If these checks fail, the 
         * <code>OptimisticLockException</code> will be thrown.
         * <p>If the lock mode type is pessimistic and the entity instance
         * is found but cannot be locked:
         * <ul>
         * <li> the <code>PessimisticLockException</code> will be thrown if the database
         *    locking failure causes transaction-level rollback
         * <li> the <code>LockTimeoutException</code> will be thrown if the database
         *    locking failure causes only statement-level rollback
         * </ul>
         * @param entity  entity instance
         * @param lockMode  lock mode
         * @throws ArgumentException if the instance is not an
         *         entity or is a detached entity
         * @throws TransactionRequiredException if there is no 
         *         transaction
         * @throws EntityNotFoundException if the entity does not exist 
         *         in the database when pessimistic locking is 
         *         performed
         * @throws OptimisticLockException if the optimistic version 
         *         check fails
         * @throws PessimisticLockException if pessimistic locking fails 
         *         and the transaction is rolled back
         * @throws LockTimeoutException if pessimistic locking fails and
         *         only the statement is rolled back
         * @throws PersistenceException if an unsupported lock call 
         *         is made
         */
        public void Lock(object entity, LockModeType lockMode)
        {
            Lock(entity, lockMode, null);
        }

        /**
         * Lock an entity instance that is contained in the persistence
         * context with the specified lock mode type and with specified
         * properties.
         * <p>If a pessimistic lock mode type is specified and the entity
         * contains a version attribute, the persistence provider must 
         * also perform optimistic version checks when obtaining the 
         * database lock.  If these checks fail, the 
         * <code>OptimisticLockException</code> will be thrown.
         * <p>If the lock mode type is pessimistic and the entity instance
         * is found but cannot be locked:
         * <ul>
         * <li> the <code>PessimisticLockException</code> will be thrown if the database
         *    locking failure causes transaction-level rollback
         * <li> the <code>LockTimeoutException</code> will be thrown if the database
         *    locking failure causes only statement-level rollback
         * </ul>
         * <p>If a vendor-specific property or hint is not recognized, 
         * it is silently ignored.  
         * <p>Portable applications should not rely on the standard timeout
         * hint. Depending on the database in use and the locking
         * mechanisms used by the provider, the hint may or may not
         * be observed.
         * @param entity  entity instance
         * @param lockMode  lock mode
         * @param properties  standard and vendor-specific properties
         *        and hints
         * @throws ArgumentException if the instance is not an
         *         entity or is a detached entity
         * @throws TransactionRequiredException if there is no 
         *         transaction
         * @throws EntityNotFoundException if the entity does not exist 
         *         in the database when pessimistic locking is 
         *         performed
         * @throws OptimisticLockException if the optimistic version 
         *         check fails
         * @throws PessimisticLockException if pessimistic locking fails 
         *         and the transaction is rolled back
         * @throws LockTimeoutException if pessimistic locking fails and
         *         only the statement is rolled back
         * @throws PersistenceException if an unsupported lock call 
         *         is made
         * @since Java Persistence 2.0
         */
        public void Lock(object entity, LockModeType lockMode,
                         Dictionary<string, object> properties)
        { 
		    try 
            {
			    if ( !isTransactionInProgress() ) 
                {
				    throw new TransactionRequiredException( "no transaction is in progress" );
			    }
			    if ( !Contains( entity ) ) 
                {
				    throw new ArgumentException( "entity not in the persistence context" );
			    }
			    getSession().Lock(entity, LockModeTypeHelper.getLockMode(lockMode));
		    }
		    catch ( HibernateException he ) {
			    throw convert(he);
		    }
        }

        /**
         * Refresh the state of the instance from the database, 
         * overwriting changes made to the entity, if any. 
         * @param entity  entity instance
         * @throws ArgumentException if the instance is not
         *         an entity or the entity is not managed
         * @throws TransactionRequiredException if invoked on a
         *         container-managed entity manager of type 
         *         <code>PersistenceContextType.TRANSACTION</code> and there is 
         *         no transaction
         * @throws EntityNotFoundException if the entity no longer
         *         exists in the database
         */
        public void Refresh(object entity)
        {
            innerRefresh(entity, null, null);
        }

        /**
         * Refresh the state of the instance from the database, using 
         * the specified properties, and overwriting changes made to
         * the entity, if any. 
         * <p> If a vendor-specific property or hint is not recognized, 
         * it is silently ignored. 
         * @param entity  entity instance
         * @param properties  standard and vendor-specific properties 
         *        and hints
         * @throws ArgumentException if the instance is not 
         *         an entity or the entity is not managed 
         * @throws TransactionRequiredException if invoked on a 
         *         container-managed entity manager of type
         *         <code>PersistenceContextType.TRANSACTION</code> and there is
         *         no transaction 
         * @throws EntityNotFoundException if the entity no longer 
         *         exists in the database 
         * @since Java Persistence 2.0
         */
        public void Refresh(object entity,
                                Dictionary<string, object> properties)
        {
            innerRefresh(entity, null, properties);
        }

        /**
         * Refresh the state of the instance from the database, 
         * overwriting changes made to the entity, if any, and 
         * lock it with respect to given lock mode type.
         * <p>If the lock mode type is pessimistic and the entity instance
         * is found but cannot be locked:
         * <ul>
         * <li> the <code>PessimisticLockException</code> will be thrown if the database
         *    locking failure causes transaction-level rollback
         * <li> the <code>LockTimeoutException</code> will be thrown if the
         *    database locking failure causes only statement-level 
         *    rollback.
         * </ul>
         * @param entity  entity instance
         * @param lockMode  lock mode
         * @throws ArgumentException if the instance is not
         *         an entity or the entity is not managed
         * @throws TransactionRequiredException if there is no 
         *         transaction and if invoked on a container-managed
         *         <code>EntityManager</code> instance with
         *         <code>PersistenceContextType.TRANSACTION</code> or with a lock mode 
         *         other than <code>NONE</code>
         * @throws EntityNotFoundException if the entity no longer exists
         *         in the database
         * @throws PessimisticLockException if pessimistic locking fails
         *         and the transaction is rolled back
         * @throws LockTimeoutException if pessimistic locking fails and
         *         only the statement is rolled back
         * @throws PersistenceException if an unsupported lock call
         *         is made
         * @since Java Persistence 2.0
         */
        public void Refresh(object entity, LockModeType lockMode)
        {
            innerRefresh(entity, lockMode, null);
        }

        /**
         * Refresh the state of the instance from the database, 
         * overwriting changes made to the entity, if any, and 
         * lock it with respect to given lock mode type and with
         * specified properties.
         * <p>If the lock mode type is pessimistic and the entity instance
         * is found but cannot be locked:
         * <ul>
         * <li> the <code>PessimisticLockException</code> will be thrown if the database
         *    locking failure causes transaction-level rollback
         * <li> the <code>LockTimeoutException</code> will be thrown if the database
         *    locking failure causes only statement-level rollback
         * </ul>
         * <p>If a vendor-specific property or hint is not recognized, 
         *    it is silently ignored.  
         * <p>Portable applications should not rely on the standard timeout
         * hint. Depending on the database in use and the locking
         * mechanisms used by the provider, the hint may or may not
         * be observed.
         * @param entity  entity instance
         * @param lockMode  lock mode
         * @param properties  standard and vendor-specific properties
         *        and hints
         * @throws ArgumentException if the instance is not
         *         an entity or the entity is not managed
         * @throws TransactionRequiredException if there is no 
         *         transaction and if invoked on a container-managed
         *         <code>EntityManager</code> instance with
         *         <code>PersistenceContextType.TRANSACTION</code> or with a lock mode 
         *         other than <code>NONE</code>
         * @throws EntityNotFoundException if the entity no longer exists
         *         in the database
         * @throws PessimisticLockException if pessimistic locking fails
         *         and the transaction is rolled back
         * @throws LockTimeoutException if pessimistic locking fails and
         *         only the statement is rolled back
         * @throws PersistenceException if an unsupported lock call
         *         is made
         * @since Java Persistence 2.0
         */
        public void Refresh(object entity, LockModeType lockModeType,
                            Dictionary<string, object> properties)
        {
            innerRefresh(entity, lockModeType, properties);
        }

        public void innerRefresh(object entity, LockModeType? lockModeType,
                            Dictionary<string, object> properties)
        {
            checkTransactionNeeded();
            CacheMode previousCacheMode = getSession().CacheMode;
            CacheMode localCacheMode = determineAppropriateLocalCacheMode(properties);
            try
            {
                getSession().CacheMode = localCacheMode;
                if (!getSession().Contains(entity))
                {
                    throw new ArgumentException("Entity not managed");
                }
                if (lockModeType.HasValue)
                {
                    LockMode lm = LockModeTypeHelper.getLockMode(lockModeType.Value);
                    getSession().Refresh(entity, lm);
                }
                else
                {
                    getSession().Refresh(entity);
                }
            }
            catch (MappingException e)
            {
                throw new ArgumentException(e.Message, e);
            }
            catch (HibernateException he)
            {
                throw convert(he);
            }
            finally
            {
                getSession().CacheMode = previousCacheMode;
            }
        }

        /**
         * Clear the persistence context, causing all managed
         * entities to become detached. Changes made to entities that
         * have not been flushed to the database will not be
         * persisted.
         */
        public void Clear()
        {
            try
            {
                getSession().Clear();
            }
            catch (HibernateException he)
            {
                throw convert(he);
            }
        }

        /**
         * Remove the given entity from the persistence context, causing
         * a managed entity to become detached.  Unflushed changes made 
         * to the entity if any (including removal of the entity), 
         * will not be synchronized to the database.  Entities which 
         * previously referenced the detached entity will continue to
         * reference it.
         * @param entity  entity instance
         * @throws ArgumentException if the instance is not an 
         *         entity 
         * @since Java Persistence 2.0
         */
        public void Detach(object entity)
        {
            try
            {
                getSession().Evict(entity);
            }
            catch (HibernateException he)
            {
                throw convert(he);
            }
        }

        /**
         * Check if the instance is a managed entity instance belonging
         * to the current persistence context.
         * @param entity  entity instance
         * @return boolean indicating if entity is in persistence context
         * @throws ArgumentException if not an entity
         */
        public bool Contains(object entity)
        {
            try 
            {
			    if ( entity != null
					    //&& !( entity is HibernateProxy )
					    && getSession().SessionFactory.GetClassMetadata( entity.GetType() ) == null ) 
                {
				    throw new ArgumentException( "Not an entity:" + entity.GetType() );
			    }
			    return getSession().Contains( entity );
		    }
		    catch ( MappingException e ) 
            {
			    throw new ArgumentException( e.Message, e );
		    }
		    catch ( HibernateException he ) 
            {
			    throw convert( he );
		    }
        }

        /**
         * Get the current lock mode for the entity instance.
         * @param entity  entity instance
         * @return lock mode
         * @throws TransactionRequiredException if there is no 
         *         transaction
         * @throws ArgumentException if the instance is not a
         *         managed entity and a transaction is active
         * @since Java Persistence 2.0
         */
        public LockModeType GetLockMode(object entity)
        {
            if (!Contains(entity))
            {
                throw new ArgumentException("entity not in the persistence context");
            }
            LockMode mode = getSession().GetCurrentLockMode(entity);
            return LockModeTypeHelper.getLockModeType(mode);
        }

        /** 
         * Set an entity manager property or hint. 
         * If a vendor-specific property or hint is not recognized, it is
         * silently ignored. 
         * @param propertyName name of property or hint
         * @param value  value for property or hint
         * @throws ArgumentException if the second argument is 
         *         not valid for the implementation 
         * @since Java Persistence 2.0
         */
        public void SetProperty(string propertyName, object value)
        {
            if (m_entityManagerSpecificProperties.Contains(propertyName))
            {
                m_properties[propertyName] = value;
                applyProperties();
            }
            else
            {
                logger.Debug("Trying to set a property which is not supported on entity manager level");
            }
        }

        private void applyProperties()
        {
            object fm;
            m_properties.TryGetValue(AvailableSettings.FLUSH_MODE, out fm);
            FlushMode? flushMode = ConfigurationHelper.getFlushMode(fm);
            if (flushMode.HasValue)
            {
                getSession().FlushMode = flushMode.Value;
            }
            else
            {
                getSession().FlushMode = default(FlushMode);
            }
            getSession().CacheMode = determineAppropriateLocalCacheMode(null);            
        }

        /**
         * Get the properties and hints and associated values that are in effect 
         * for the entity manager. Changing the contents of the map does 
         * not change the configuration in effect.
         * @return map of properties and hints in effect for entity manager
         * @since Java Persistence 2.0
         */
        public Dictionary<string, object> GetProperties()
        {
            Dictionary<string, object> props = new Dictionary<string, object>();
            foreach (KeyValuePair<string, object> property in m_properties)
            {
                props.Add((string)property.Key, property.Value);
            }
            return props;
        }

        /**
         * Create an instance of <code>Query</code> for executing a
         * Java Persistence query language statement.
         * @param qlString a Java Persistence query string
         * @return the new query instance
         * @throws ArgumentException if the query string is
         *         found to be invalid
         */
        public Query CreateQuery(string qlString)
        {
            IQuery hibernateQuery = m_session.CreateQuery(qlString);
            return new QueryImpl<object>(hibernateQuery, this);
            
        }

        /**
         * Create an instance of <code>TypedQuery</code> for executing a
         * criteria query.
         * @param criteriaQuery  a criteria query object
         * @return the new query instance
         * @throws ArgumentException if the criteria query is
         *         found to be invalid
         * @since Java Persistence 2.0
         */
        //    TypedQuery<T> CreateQuery<T> (CriteriaQuery<T> criteriaQuery); 

        /**
         * Create an instance of <code>TypedQuery</code> for executing a
         * Java Persistence query language statement.
         * The select list of the query must contain only a single
         * item, which must be assignable to the type specified by
         * the <code>resultClass</code> argument.
         * @param qlString a Java Persistence query string
         * @param resultClass the type of the query result
         * @return the new query instance
         * @throws ArgumentException if the query string is found
         *         to be invalid or if the query result is found to
         *         not be assignable to the specified type
         * @since Java Persistence 2.0
         */
        public TypedQuery<T> CreateQuery<T>(string qlString, System.Type resultClass)
        {
            try
            {
                IQuery hqlQuery = getSession().CreateQuery(qlString);
                if (hqlQuery.ReturnTypes.Length != 1)
                {
                    throw new ArgumentException("Cannot create TypedQuery for query with more than one return");
                }
                if (!resultClass.IsAssignableFrom(hqlQuery.ReturnTypes[0].ReturnedClass))
                {
                    throw new ArgumentException(
                            "Type specified for TypedQuery [" +
                                    resultClass.Name +
                                    "] is incompatible with query return type [" +
                                    hqlQuery.ReturnTypes[0].ReturnedClass + "]"
                    );
                }
                return new QueryImpl<T>(hqlQuery, this);
            }
            catch (HibernateException he)
            {
                throw convert(he);
            }
        }

        /**
         * Create an instance of <code>Query</code> for executing a named query
         * (in the Java Persistence query language or in native SQL).
         * @param name the name of a query defined in metadata
         * @return the new query instance
         * @throws ArgumentException if a query has not been
         *         defined with the given name or if the query string is
         *         found to be invalid
         */
        public Query CreateNamedQuery(string name)
        {
            try
            {
                IQuery namedQuery = getSession().GetNamedQuery(name);
                try
                {
                    return new QueryImpl<object>(namedQuery, this);
                }
                catch (HibernateException he)
                {
                    throw convert(he);
                }
            }
            catch (MappingException e)
            {
                throw new ArgumentException("Named query not found: " + name);
            }
        }

        /**
         * Create an instance of <code>TypedQuery</code> for executing a
         * Java Persistence query language named query.
         * The select list of the query must contain only a single
         * item, which must be assignable to the type specified by
         * the <code>resultClass</code> argument.
         * @param name the name of a query defined in metadata
         * @param resultClass the type of the query result
         * @return the new query instance
         * @throws ArgumentException if a query has not been
         *         defined with the given name or if the query string is
         *         found to be invalid or if the query result is found to
         *         not be assignable to the specified type
         * @since Java Persistence 2.0
         */
        public TypedQuery<T> CreateNamedQuery<T>(string name, System.Type resultClass)
        {
            IQuery namedQuery;
            try
            {
                namedQuery = getSession().GetNamedQuery(name);
            }
            catch (MappingException e)
            {
                throw new ArgumentException("Named query not found: " + name);
            }
            try
            {
                return new QueryImpl<T>(namedQuery, this);
            }
            catch (HibernateException he)
            {
                throw convert(he);
            }
        }


        /**
         * Create an instance of <code>Query</code> for executing
         * a native SQL statement, e.g., for update or delete.
         * @param sqlString a native SQL query string
         * @return the new query instance
         */
        public Query CreateNativeQuery(string sqlString)
        {
            try
            {
                ISQLQuery q = getSession().CreateSQLQuery(sqlString);
                return new QueryImpl<object>(q, this);
            }
            catch (HibernateException he)
            {
                throwPersistenceException(he);
                return null;
            }
        }


        /**
         * Create an instance of <code>Query</code> for executing
         * a native SQL query.
         * @param sqlString a native SQL query string
         * @param resultClass the class of the resulting instance(s)
         * @return the new query instance
         */
        public Query CreateNativeQuery(string sqlString, System.Type resultClass)
        {
            try
            {
                ISQLQuery q = getSession().CreateSQLQuery(sqlString);
                q.AddEntity("alias1", resultClass, LockMode.Read);
                return new QueryImpl<object>(q, this);
            }
            catch (HibernateException he)
            {
                throw convert(he);
            }
        }


        /**
         * Create an instance of <code>Query</code> for executing
         * a native SQL query.
         * @param sqlString a native SQL query string
         * @param resultSetMapping the name of the result set mapping
         * @return the new query instance
         */
        public Query CreateNativeQuery(string sqlString, string resultSetMapping)
        {
            try
            {
                ISQLQuery q = getSession().CreateSQLQuery(sqlString);
                q.SetResultSetMapping(resultSetMapping);
                return new QueryImpl<object>(q, this);
            }
            catch (HibernateException he)
            {
                throw convert(he);
            }
        }


        /**
         * Indicate to the entity manager that a JTA transaction is
         * active. This method should be called on a JTA application 
         * managed entity manager that was created outside the scope
         * of the active transaction to associate it with the current
         * JTA transaction.
         * @throws TransactionRequiredException if there is
         *         no transaction
         */
        public void JoinTransaction()
        {
            throw new NotSupportedException("JTA is not supported");
	    }

        /**
         * Return an object of the specified type to allow access to the
         * provider-specific API.   If the provider's <code>EntityManager</code>
         * implementation does not support the specified class, the
         * <code>PersistenceException</code> is thrown.
         * @param cls  the class of the object to be returned.  This is
         * normally either the underlying <code>EntityManager</code> implementation
         * class or an interface that it implements.
         * @return an instance of the specified class
         * @throws PersistenceException if the provider does not 
         *         support the call 
         * @since Java Persistence 2.0
         */
        public T Unwrap<T>(System.Type clazz)
        {
            if ( typeof(ISession).IsAssignableFrom( clazz ) ) 
            {
			    return ( T ) getSession();
		    }
		    if ( typeof(ISessionImplementor).IsAssignableFrom( clazz ) ) 
            {
			    return ( T ) getSession();
		    }
		    if ( typeof(EntityManager).IsAssignableFrom( clazz ) ) 
            {
			    return ( T ) ((EntityManager)this);
		    }
		    throw new PersistenceException( "Hibernate cannot unwrap " + clazz );
 
        }

        /**
         * Return the underlying provider object for the <code>EntityManager</code>,
         * if available. The result of this method is implementation
         * specific. The <code>unwrap</code> method is to be preferred for new 
         * applications.
         * @return underlying provider object for EntityManager
         */
        public object GetDelegate()
        {
            return getSession();
        }

        private class CloseSynch : ISynchronization
        {
            private ILog logger;
            private global::NHibernate.ISession m_sessionToClose;

            public CloseSynch(ISession sessionToClose, ILog logger)
            {
                this.m_sessionToClose = sessionToClose;
                this.logger = logger;
            }
            
            public void AfterCompletion(bool success)
            {
                if (m_sessionToClose != null)
				{
					if (m_sessionToClose.IsOpen)
					{
						logger.Debug("Closing entity manager after transaction completion");
						m_sessionToClose.Close();
					} 
                    else
					{
						logger.Warn("Entity Manager closed by someone else ");

					}
				}
            }
            public void BeforeCompletion()
            {
            }
        }
        /**
         * Close an application-managed entity manager. 
         * After the close method has been invoked, all methods
         * on the <code>EntityManager</code> instance and any 
         * <code>Query</code> and <code>TypedQuery</code>
         * objects obtained from it will throw the <code>InvalidOperationException</code>
         * except for <code>getProperties</code>, 
         * <code>getTransaction</code>, and <code>isOpen</code> (which will return false).
         * If this method is called when the entity manager is
         * associated with an active transaction, the persistence
         * context remains managed until the transaction completes. 
         * @throws InvalidOperationException if the entity manager
         *         is container-managed
         */
        public void Close()
        {
            if (!m_open)
		    {
                throw new InvalidOperationException("EntityManager is closed");
		    }
		    if (!m_discardOnClose && isTransactionInProgress())
		    {
			    // delay the closing till the end of the enlisted transaction
                getSession().Transaction.RegisterSynchronization(new CloseSynch(m_session, logger));
					    
		    } 
            else
		    {
			    // close right now
			    if (m_session != null)
			    {
				    m_session.Close();
			    }
		    }
            m_open = false;
        }

        /**
         * Determine whether the entity manager is open. 
         * @return true until the entity manager has been closed
         */
        public bool IsOpen()
        {
            return m_open;
        }

        /**
         * Return the resource-level <code>EntityTransaction</code> object. 
         * The <code>EntityTransaction</code> instance may be used serially to 
         * begin and commit multiple transactions.
         * @return EntityTransaction instance
         * @throws InvalidOperationException if invoked on a JTA
         *         entity manager
         */
        public EntityTransaction GetTransaction()
        {
            //if (m_transactionType == PersistenceUnitTransactionType.JTA)
            //{
            //    throw new StateException("A JTA EntityManager cannot use getTransaction()");
            //}
            return m_tx;
        }

        /**
         * Return the entity manager factory for the entity manager.
         * @return EntityManagerFactory instance
         * @throws InvalidOperationException if the entity manager has 
         *         been closed
         * @since Java Persistence 2.0
         */
        public EntityManagerFactory GetEntityManagerFactory()
        {
            return m_entityManagerFactory;
        }
    }
}
