﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using NPersistence;
using log4net;
using NHibernate.Util;
using NHibernate;
using NHibernate.Cfg;
using System.Reflection;
using NPersistence.NHibernate.packaging;
using NPersistence.spi;
using System.IO;
using System.Threading;
using NPersistence.NHibernate.annotationSupport;
using System.Xml.Serialization;
using NPersistence.NHibernate.annotationSupport.mapping;
using System.Collections.Specialized;

namespace NPersistence.NHibernate
{
    public class NHibernateConfiguration : Configuration
    {
        private ILog logger = LogManager.GetLogger("NPersistence.NHibernate");

        private const String IMPLEMENTATION_NAME = "NPersistence.NHibernate.NHibernatePersistence";//System.Type.GetType("NHibernatePersistence").Name;

        private PersistenceUnitTransactionType m_transactionType;

        private bool m_discardOnClose;
        private string m_persistenceUnitName = null;
        /**
	    * Build the configuration from an entity manager name and given the
	    * appropriate extra properties. Those properties override the one get through
	    * the peristence.xml file.
	    * If the persistence unit name is not found or does not match the Persistence Provider, null is returned
	    *
	    * This method is used in a non managed environment
	    *
	    * @param persistenceUnitName persistence unit name
	    * @param properties properties passed to the persistence provider
	    * @return configured Ejb3Configuration or null if no persistence unit match
	    */
        public NHibernateConfiguration Configure(string persistenceUnitName, Dictionary<object, object> properties)
        { 
            try 
            {
                if(logger.IsDebugEnabled)
                {
                    logger.Debug( "Look up for persistence unit: " + persistenceUnitName );
                }
			
                if(properties == null)
                {
                    properties = new Dictionary<object, object>();
                }

                // First try to get configuration from 'web.config' file.
                List<PersistenceMetadata> metadataList = PersistenceXmlLoader.webDeploy(properties);
                NHibernateConfiguration config = null;
                if (metadataList != null && metadataList.Count > 0)
                {
                    config = configure(persistenceUnitName, properties, metadataList);
                    if (config != null)
                    {
                        return config;
                    }
                }
                
                // Find the persistence.xml file.
                bool persistenceXmlFound = false;
                List<Assembly> assemblies = new List<Assembly>();
                Assembly loader = Assembly.GetEntryAssembly();
                if(loader != null)
                {
                    assemblies.Add(loader);
                }

                Assembly[] ass = AppDomain.CurrentDomain.GetAssemblies();
                assemblies.AddRange(ass);

                foreach (Assembly assembly in assemblies)
                {
                    String presistenceXmlResourceName = getPersistenceResource(assembly);
                    if (presistenceXmlResourceName != null)
                    {
                        persistenceXmlFound = true;
                        System.IO.Stream stream = assembly.GetManifestResourceStream(presistenceXmlResourceName);
                        if (logger.IsDebugEnabled)
                        {
                            logger.Debug("Analysing persistence.xml");
                        }
                        metadataList = PersistenceXmlLoader.deploy(stream, properties);
                        config = configure(persistenceUnitName, properties, metadataList);
                        if(config != null)
                        {
                            return config;
                        }
                    }
                }
                
			    if ( !persistenceXmlFound) 
                {
                    if(logger.IsInfoEnabled)
                    {
    				    logger.Info( "Could not find any persistence.xml file in the classpath");
                    }   
			    }
                return null;
		    }
		    catch (Exception e) 
            {
			    if ( e is PersistenceException) 
                {
				    throw (PersistenceException) e;
			    }
			    else 
                {
				    throw new PersistenceException("Unable to configure EntityManagerFactory", e );
			    }
		    }
        }

        private NHibernateConfiguration configure(string persistenceUnitName, Dictionary<object, object> properties, List<PersistenceMetadata> metadataList)
        {
           
            foreach (PersistenceMetadata metadata in metadataList)
            {
                if (logger.IsDebugEnabled)
                {
                    logger.Debug(metadata);
                }

                if (metadata.getProvider() == null ||
                    IMPLEMENTATION_NAME.ToLower().Equals(metadata.getProvider().ToLower()))
                {
                    //correct provider

                    if (persistenceUnitName == null || metadata.getName().Equals(persistenceUnitName))
                    {
                        return configure(metadata);
                    }
                }
            }
            return null;
        }

        private static string getPersistenceResource(Assembly loader)
        {
            string[] resources = null;
            try
            {
                resources = loader.GetManifestResourceNames();
            }
            catch (SystemException e)
            {
                return null;
            }
            if (resources == null)
            {
                return null;
            }
            string presistenceXmlResourceName = null;
            foreach (string resourceName in resources)
            {
                if (resourceName.EndsWith("META-INF.persistence.xml"))
                {
                    presistenceXmlResourceName = resourceName;
                    break;
                }
                else if (resourceName.EndsWith("persistence.xml"))
                {
                    presistenceXmlResourceName = resourceName;
                }
            }
            return presistenceXmlResourceName;
        }

        private NHibernateConfiguration configure(PersistenceMetadata metadata)
        {
            this.SetProperties(metadata.getProps());
            Dictionary<string, string> properties =  metadata.getProps();
            string discardOnClose;
            properties.TryGetValue(AvailableSettings.DISCARD_PC_ON_CLOSE, out discardOnClose);
            m_discardOnClose = discardOnClose != null && discardOnClose.ToLower().Equals("true");
            m_transactionType = metadata.getTransactionType();
            m_persistenceUnitName = metadata.getName();

            foreach (string file in metadata.getMappingFiles())
            {
                this.AddFile(file);
            }
            foreach (string assemblyFile in metadata.getJarFiles())
            {
                this.AddAssembly(assemblyFile);
            }
            
            AnnotaionConfig annotationConfig = AnnotaionConfig.getInstance(properties);
            annotationConfig.configure(metadata.getClasses(), this);
            return this;
        }
        
        public EntityManagerFactory BuildEntityManagerFactory()
        {
            try
            {
                ISessionFactory sessionFactory = BuildSessionFactory();
                Type interceptor = getSessionInterceptorClass();
                return new EntityManagerFactoryImpl(sessionFactory, m_transactionType, interceptor, m_discardOnClose);
            }
            catch (Exception e)
            {
                if (e is PersistenceException)
                {
                    throw (PersistenceException)e;
                }
                else
                {
                    throw new PersistenceException("Unable to build EntityManagerFactory", e);
                }
            }
        }

        private Type getSessionInterceptorClass() 
        {
		    String sessionInterceptorClassname = null;
            Properties.TryGetValue(AvailableSettings.SESSION_INTERCEPTOR, out sessionInterceptorClassname);
		    if ( StringHelper.IsNotEmpty( sessionInterceptorClassname ) ) 
            {
			    try 
                {
				    Type interceptorClass = ReflectHelper.ClassForName( sessionInterceptorClassname);
                    if (interceptorClass == null)
                    {
                        throw new PersistenceException(getExceptionHeader() + "Unable to load "
                            + AvailableSettings.SESSION_INTERCEPTOR + ": " + sessionInterceptorClassname);
                    }
                    ConstructorInfo constructor = interceptorClass.GetConstructor(System.Type.EmptyTypes);
                    if (constructor == null)
                    {
                        throw new PersistenceException(getExceptionHeader() + "Unable to instanciate "
                            + AvailableSettings.SESSION_INTERCEPTOR + ": " + sessionInterceptorClassname + ". No empty constructor found");
                    }

                    constructor.Invoke(System.Type.EmptyTypes);
				    return interceptorClass;
			    }

			    catch (MethodAccessException e) 
                {
				    throw new PersistenceException( getExceptionHeader() + "Unable to instanciate "
						    + AvailableSettings.SESSION_INTERCEPTOR + ": " + sessionInterceptorClassname, e);
			    }
                catch (MemberAccessException e)
                {
                    throw new PersistenceException(getExceptionHeader() + "Unable to instanciate "
                            + AvailableSettings.SESSION_INTERCEPTOR + ": " + sessionInterceptorClassname, e);
                }
			    catch (InstantiationException e) 
                {
				    throw new PersistenceException( getExceptionHeader() + "Unable to instanciate "
						    + AvailableSettings.SESSION_INTERCEPTOR + ": " + sessionInterceptorClassname, e);
			    }
		    }
		    else {
			    return null;
		    }
	    }

        private String getExceptionHeader()
        {
            if (StringHelper.IsNotEmpty(m_persistenceUnitName))
            {
                return "[PersistenceUnit: " + m_persistenceUnitName + "] ";
            }
            else
            {
                return "";
            }
        }
    }
}
