﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using NPersistence;
using System.Collections;
using NHibernate;
using NHibernate.Impl;
using System.Globalization;
using NPersistence.NHibernate.util;
using log4net;
using NPersistence.NHibernate;
using NHibernate.Hql;

namespace NPersistence.NHibernate.query
{
    class QueryImpl<X> : TypedQuery<X>
    {
        private ILog logger = LogManager.GetLogger("NPersistence.NHibernate");
        
        /**
	     * Query hint (aka {@link javax.persistence.Query#setHint}) for applying
	     * an alias specific lock mode (aka {@link org.hibernate.Query#setLockMode}).
	     * <p/>
	     * Either {@link org.hibernate.LockMode} or {@link javax.persistence.LockModeType}
	     * are accepted.  Also the String names of either are accepted as well.  <tt>null</tt>
	     * is additionally accepted as meaning {@link org.hibernate.LockMode#NONE}.
	     * <p/>
	     * Usage is to concatenate this setting name and the alias name together, separated
	     * by a dot.  For example<code>Query.setHint( "org.hibernate.lockMode.a", someLockMode )</code>
	     * would apply <code>someLockMode</code> to the alias <code>"a"</code>.
	     */
	    //Use the org.hibernate prefix. instead of hibernate. as it is a query hint se QueryHints
	    public const String ALIAS_SPECIFIC_LOCK_MODE = "org.hibernate.lockMode";

        private IQuery m_hibernateQuery;
        private EntityManagerImpl m_entityManager;

        public static readonly int NoValue = -1;
        private int m_maxResults = NoValue;
        private int m_firstResult = NoValue;
        private FlushModeType? m_flushMode = null;
        private LockModeType m_lockMode = LockModeType.NONE;

        private Dictionary<Parameter<Object>, Object> m_parameterBindings;
        HashSet<Object> m_parameters = new HashSet<Object>();

        private Dictionary<String, Object> m_hints;

        public QueryImpl(IQuery hibernateQuery, EntityManagerImpl entityManager)
        {
            this.m_hibernateQuery = hibernateQuery;
            this.m_entityManager = entityManager;
        }


        IList<X> TypedQuery<X>.GetResultList()
        {
            return GetResultList();
        }

        IList Query.GetResultList()
        {
            return (IList)GetResultList();
        }

        private IList<X> GetResultList()
        {
            try
            {
                return (IList<X>)m_hibernateQuery.List<X>();
            }
            catch (QueryExecutionRequestException he)
            {
                throw new InvalidOperationException(he.Message, he);
            }
            catch (TypeMismatchException e)
            {
                throw new ArgumentException(e.Message, e);
            }
            catch (HibernateException he)
            {
                throw m_entityManager.convert(he);
            }
        }

        X TypedQuery<X>.GetSingleResult()
        {
            return GetSingleResult();
        }

        Object Query.GetSingleResult()
        {
            return GetSingleResult();
        }

        private X GetSingleResult()
        {
            try
            {
                if (m_maxResults != 1)
                {
                    m_hibernateQuery.SetMaxResults(2); //avoid OOME if the list is huge
                }
                IList result = m_hibernateQuery.List();

                if (result.Count == 0)
                {
                    NoResultException nre = new NoResultException("No entity found for query");
                    m_entityManager.handlePersistenceException(nre);
                    throw nre;
                }
                else if (result.Count > 1)
                {
                    NonUniqueResultException nure = new NonUniqueResultException("result returns more than one elements");
                    m_entityManager.handlePersistenceException(nure);
                    throw nure;
                }
                else
                {
                    return (X)result[0];
                }
            }
            catch (QueryExecutionRequestException he)
            {
                throw new InvalidOperationException(he.Message, he);
            }
            catch (TypeMismatchException e)
            {
                throw new ArgumentException(e.Message, e);
            }
            catch (HibernateException he)
            {
                throw m_entityManager.convert(he);
            }
            finally
            {
                if (m_maxResults != 1)
                {
                    m_hibernateQuery.SetMaxResults(m_maxResults);
                }

            }
        }

        public int ExecuteUpdate()
        {
            return m_hibernateQuery.ExecuteUpdate();
        }

        TypedQuery<X> TypedQuery<X>.SetMaxResults(int maxResult)
        {
            return SetMaxResults(maxResult);
        }

        Query Query.SetMaxResults(int maxResult)
        {
            return SetMaxResults(maxResult);
        }
        
        private TypedQuery<X> SetMaxResults(int maxResult)
        {
            m_hibernateQuery.SetMaxResults(maxResult);
            m_maxResults = maxResult;
            return this;
        }

        // <remarks>Defaults to NoValue unless specifically set.</remarks>
        public int GetMaxResults()
        { 
            if(m_hibernateQuery is AbstractQueryImpl)
            {
                AbstractQueryImpl imp = (AbstractQueryImpl)m_hibernateQuery;
                return imp.Selection.MaxRows;
            }
            return m_maxResults;
        }

        /**
         * Set the position of the first result to retrieve.
         * @param startPosition position of the first result, 
         * numbered from 0
         * @return the same query instance
         * @throws ArgumentException if the argument is negative
         */
        Query Query.SetFirstResult(int startPosition)
        {
            return SetFirstResult(startPosition);
        }

        TypedQuery<X> TypedQuery<X>.SetFirstResult(int startPosition)
        {
            return SetFirstResult(startPosition);
        }

        private TypedQuery<X> SetFirstResult(int startPosition)
        {
            m_hibernateQuery.SetFirstResult(startPosition);
            return this;
        }

        /**
         * The position of the first result the query object was set to
         * retrieve. Returns 0 if <code>setFirstResult</code> was not applied to the
         * query object.
         * @return position of the first result
         * @since Java Persistence 2.0
         */
        public int GetFirstResult()
        { 
            if(m_hibernateQuery is AbstractQueryImpl)
            {
                AbstractQueryImpl imp = (AbstractQueryImpl)m_hibernateQuery;
                return imp.Selection.FirstRow;
            }
            return m_firstResult;
        }

        
        /**
         * Set a query property or hint. The hints elements may be used 
         * to specify query properties and hints. Properties defined by
         * this specification must be observed by the provider. 
         * Vendor-specific hints that are not recognized by a provider
         * must be silently ignored. Portable applications should not
         * rely on the standard timeout hint. Depending on the database
         * in use and the locking mechanisms used by the provider,
         * this hint may or may not be observed.
         * @param hintName  name of the property or hint
         * @param value  value for the property or hint
         * @return the same query instance
         * @throws ArgumentException if the second argument is not
         *         valid for the implementation
         */
        TypedQuery<X> TypedQuery<X>.SetHint(string hintName, object value)
        {
            return SetHint(hintName, value);
        }

        Query Query.SetHint(string hintName, object value)
        {
            return SetHint(hintName, value);
        }

        private TypedQuery<X> SetHint(string hintName, object value)
        {
            bool skipped = false;
		    try {
   			    if ( QueryHints.HINT_TIMEOUT.Equals( hintName ) ) 
             {
				    m_hibernateQuery.SetTimeout(ConfigurationHelper.getInteger(value));
			    }
                else if (QueryHints.SPEC_HINT_TIMEOUT.Equals(hintName) || QueryHints.SPEC_HINT_TIMEOUT2.Equals(hintName)) 
                {
    				// convert milliseconds to seconds
                    double millis = ConfigurationHelper.getInteger(value);
				    int timeout = (int)Math.Round( millis / 1000.0 ); 
				    m_hibernateQuery.SetTimeout(timeout);
			    }
			    else if ( QueryHints.HINT_COMMENT.Equals( hintName ) ) 
                {
    				m_hibernateQuery.SetComment( (String) value );
			    }
			    else if ( QueryHints.HINT_FETCH_SIZE.Equals( hintName ) ) 
                {
    				m_hibernateQuery.SetFetchSize(ConfigurationHelper.getInteger(value));
			    }
			    else if ( QueryHints.HINT_CACHEABLE.Equals( hintName ) ) 
                {
    				m_hibernateQuery.SetCacheable( ConfigurationHelper.getBoolean( value ) );
			    }
			    else if ( QueryHints.HINT_CACHE_REGION.Equals( hintName ) ) 
                {
    				m_hibernateQuery.SetCacheRegion( (String) value );
			    }
			    else if ( QueryHints.HINT_READONLY.Equals( hintName ) ) 
                {
    				m_hibernateQuery.SetReadOnly( ConfigurationHelper.getBoolean( value ) );
			    }
			    else if ( QueryHints.HINT_CACHE_MODE.Equals( hintName ) ) 
                {
    				m_hibernateQuery.SetCacheMode( ConfigurationHelper.getCacheMode( value ) );
			    }
			    else if ( QueryHints.HINT_FLUSH_MODE.Equals( hintName ) ) 
                {
                    FlushMode? mode = ConfigurationHelper.getFlushMode( value );
                    if(mode.HasValue)
                    {
    				    m_hibernateQuery.SetFlushMode(mode.Value);
                    }
			    }
			    else if ( AvailableSettings.SHARED_CACHE_RETRIEVE_MODE_JPA.Equals( hintName ) || AvailableSettings.SHARED_CACHE_RETRIEVE_MODE.Equals(hintName))
                {
                    CacheRetrieveMode retrieveMode = ConfigurationHelper.getCacheRetrieveMode(value);
    
				    object storeMode = null; 
                    if(m_hints != null)
                    {
                        if (m_hints.ContainsKey(AvailableSettings.SHARED_CACHE_STORE_MODE_JPA))
                        {
                            storeMode = m_hints[AvailableSettings.SHARED_CACHE_STORE_MODE_JPA];
                        }
                        if (storeMode == null && m_hints.ContainsKey(AvailableSettings.SHARED_CACHE_STORE_MODE))
                        {
                            storeMode = m_hints[AvailableSettings.SHARED_CACHE_STORE_MODE];
                        }
                    }
				    if ( storeMode == null ) 
                    {
                        Dictionary<String, Object> emProps = m_entityManager.GetProperties();
                        if (emProps.ContainsKey(AvailableSettings.SHARED_CACHE_STORE_MODE_JPA))
                        {
                            storeMode = emProps[AvailableSettings.SHARED_CACHE_STORE_MODE_JPA];
                        }
                        if (storeMode == null && emProps.ContainsKey(AvailableSettings.SHARED_CACHE_STORE_MODE)) 
                        {
                            storeMode = emProps[AvailableSettings.SHARED_CACHE_STORE_MODE];
				        }
				    }
                    CacheStoreMode? cStoreMode = null;
                    if (storeMode != null)
                    {
                        cStoreMode = ConfigurationHelper.getCacheStoreMode(storeMode);
                    }
                    m_hibernateQuery.SetCacheMode(CacheModeHelper.interpretCacheMode( cStoreMode, retrieveMode ));
			    }
                else if (AvailableSettings.SHARED_CACHE_STORE_MODE_JPA.Equals(hintName) || AvailableSettings.SHARED_CACHE_STORE_MODE.Equals(hintName)) 
                {
    				CacheStoreMode storeMode = ConfigurationHelper.getCacheStoreMode(value);
    
				    object retrieveMode = null;
                    if(m_hints != null)
                    {
                        if (m_hints.ContainsKey(AvailableSettings.SHARED_CACHE_RETRIEVE_MODE_JPA))
                        {
                            retrieveMode = m_hints[AvailableSettings.SHARED_CACHE_RETRIEVE_MODE_JPA];
                        }
                        if (retrieveMode == null && m_hints.ContainsKey(AvailableSettings.SHARED_CACHE_RETRIEVE_MODE))
                        {
                            retrieveMode = m_hints[AvailableSettings.SHARED_CACHE_RETRIEVE_MODE];
                        }
                    }
						
				    if ( retrieveMode == null ) 
                    {
                        Dictionary<String, Object> emProps = m_entityManager.GetProperties();
                        if (emProps.ContainsKey(AvailableSettings.SHARED_CACHE_RETRIEVE_MODE_JPA))
                        {
                            retrieveMode = emProps[AvailableSettings.SHARED_CACHE_RETRIEVE_MODE_JPA];
                        }
                        if (retrieveMode == null && emProps.ContainsKey(AvailableSettings.SHARED_CACHE_RETRIEVE_MODE))
                        {
                            retrieveMode = emProps[AvailableSettings.SHARED_CACHE_RETRIEVE_MODE];
                        }
				    }
                    CacheRetrieveMode? cRetrieveMode = null;
                    if (retrieveMode != null)
                    {
                        cRetrieveMode = ConfigurationHelper.getCacheRetrieveMode(retrieveMode);
                    }
                    m_hibernateQuery.SetCacheMode(CacheModeHelper.interpretCacheMode( storeMode, cRetrieveMode ));
				
			    }
                else if (ALIAS_SPECIFIC_LOCK_MODE.StartsWith(hintName))
                {
                    throw new NotImplementedException("not implemented yet" + ALIAS_SPECIFIC_LOCK_MODE);
                }
                else
                {
                    skipped = true;
                    logger.Info("Ignoring unrecognized query hint [" + hintName + "]");
                }
		    }
		    catch ( InvalidCastException e ) 
            {
    			throw new ArgumentException( "Value for hint" );
		    }

		    if ( !skipped ) 
            {
	    		if ( m_hints == null ) 
                {
				    m_hints = new Dictionary<String,Object>();
			    }   
			    m_hints[hintName] = value;
		    }
            return this;
        }

        /**
         * Get the properties and hints and associated values that are 
         * in effect for the query instance.
         * @return query properties and hints
         * @since Java Persistence 2.0
         */
        public Dictionary<string, object> GetHints()
        { 
            return m_hints;
        }

        /**
         * Bind the value of a <code>Parameter</code> object.
         * @param param  parameter object
         * @param value  parameter value
         * @return the same query instance
         * @throws ArgumentException if the parameter
         *         does not correspond to a parameter of the
         *         query
         * @since Java Persistence 2.0
         */
        TypedQuery<X> TypedQuery<X>.SetParameter<T>(Parameter<T> param, T value)
        {
            return SetParameter(param, value);
        }

        Query Query.SetParameter<T>(Parameter<T> param, T value)
        {
            return SetParameter(param, value);
        }

        private TypedQuery<X> SetParameter<T>(Parameter<T> param, T value)
        {
            if (param.GetName() != null)
            {
                m_hibernateQuery.SetParameter<T>(param.GetName(), value);
            }
            else if (param.GetPosition().HasValue)
            {
                m_hibernateQuery.SetParameter<T>(param.GetPosition().Value, value);
            }
            else
            {
                throw new ArgumentException("The given parameter has no name nor position");
            }
            registerParameterBinding(param, value);
            return this;
        }


        /**
         * Bind an instance of <code>java.util.Calendar</code> to a <code>Parameter</code> object.
         * @param param parameter object
         * @param value  parameter value
         * @param temporalType  temporal type
         * @return the same query instance
         * @throws ArgumentException if the parameter does not
         *         correspond to a parameter of the query
         * @since Java Persistence 2.0
         */
        TypedQuery<X> TypedQuery<X>.SetParameter(Parameter<System.DateTime> param, System.DateTime value, TemporalType temporalType)
        {
            return SetParameter(param, value, temporalType);
        }

        Query Query.SetParameter(Parameter<System.DateTime> param, System.DateTime value, TemporalType temporalType)
        {
            return SetParameter(param, value, temporalType);
        }

        private TypedQuery<X> SetParameter(Parameter<System.DateTime> param, System.DateTime value, TemporalType temporalType)
        {
            if (param.GetName() != null)
            {
                SetParameter(param.GetName(), value, temporalType);
            }
            else if (param.GetPosition().HasValue)
            {
                SetParameter(param.GetPosition().Value, value, temporalType);
            }
            else
            {
                throw new ArgumentException("The given parameter has no name nor position");
            }
            registerParameterBinding(param, value);
            return this;
        }

        /**
         * Bind an argument to a named parameter.
         * @param name  parameter name
         * @param value  parameter value
         * @return the same query instance
         * @throws ArgumentException if the parameter name does 
         *         not correspond to a parameter of the query or if
         *         the argument is of incorrect type
         */
        TypedQuery<X> TypedQuery<X>.SetParameter(string name, object value)
        {
            return SetParameter(name, value);
        }

        Query Query.SetParameter(string name, object value)
        {
            return SetParameter(name, value);
        }

        private TypedQuery<X> SetParameter(string name, object value)
        {
            try
            {
                m_hibernateQuery.SetParameter(name, value);
                Parameter<Object> p = new ParameterImpl<Object>(name, value.GetType());
                registerParameterBinding(p, value);
                return this;
            }
            catch (QueryParameterException e)
            {
                throw new ArgumentException(e.Message, e);
            }
            catch (HibernateException he)
            {
                throw m_entityManager.convert(he);
            }
        }


        /**
         * Bind an instance of <code>java.util.Date</code> to a named parameter.
         * @param name  parameter name
         * @param value  parameter value
         * @param temporalType  temporal type
         * @return the same query instance
         * @throws ArgumentException if the parameter name does 
         *         not correspond to a parameter of the query or if
         *         the value argument is of incorrect type
         */
        TypedQuery<X> TypedQuery<X>.SetParameter(string name, System.DateTime value,
                           TemporalType temporalType)
        {
            return SetParameter(name, value, temporalType);
        }

        Query Query.SetParameter(string name, System.DateTime value,
                           TemporalType temporalType)
        {
            return SetParameter(name, value, temporalType);
        }

        private TypedQuery<X> SetParameter(string name, System.DateTime value,
                           TemporalType temporalType)
        { 
            try
            {
                if (temporalType == TemporalType.DATE)
                {
                    m_hibernateQuery.SetDateTime(name, value);    
                }
                else if (temporalType == TemporalType.TIME)
                {
                    m_hibernateQuery.SetTime(name, value);
                }
                else if (temporalType == TemporalType.TIMESTAMP)
                {
                    m_hibernateQuery.SetTimestamp(name, value);
                }
                Parameter<Object> p = new ParameterImpl<Object>(name, value.GetType());
                registerParameterBinding(p, value);
                return this;
            }
		    catch (QueryParameterException e) 
            {
			    throw new ArgumentException(e.Message, e);
		    }
		    catch (HibernateException he) 
            {
			    throw m_entityManager.convert(he);
		    }
        }

        /**
         * Bind an argument to a positional parameter.
         * @param position  position
         * @param value  parameter value
         * @return the same query instance
         * @throws ArgumentException if position does not
         *         correspond to a positional parameter of the
         *         query or if the argument is of incorrect type
         */
        TypedQuery<X> TypedQuery<X>.SetParameter(int position, object value)
        {
            return SetParameter(position, value);
        }
        Query Query.SetParameter(int position, object value)
        {
            return SetParameter(position, value);
        }
        private TypedQuery<X> SetParameter(int position, object value)
        {
            try 
            {
                m_hibernateQuery.SetParameter(position, value);
                Parameter<Object> p = new ParameterImpl<Object>(position, value.GetType());
                registerParameterBinding(p, value);
                return this;
            }
            catch (QueryParameterException e)
            {
                throw new ArgumentException(e.Message, e);
            }
            catch (HibernateException he)
            {
                throw m_entityManager.convert(he);
            }
        }

     
        /**
         * Bind an instance of <code>java.util.Date</code> to a positional parameter.
         * @param position  position
         * @param value  parameter value
         * @param temporalType  temporal type
         * @return the same query instance
         * @throws ArgumentException if position does not
         *         correspond to a positional parameter of the query or
         *         if the value argument is of incorrect type
         */
        TypedQuery<X> TypedQuery<X>.SetParameter(int position, System.DateTime value,
                           TemporalType temporalType)
        {
            return SetParameter(position, value, temporalType);
        }

        Query Query.SetParameter(int position, System.DateTime value,
                           TemporalType temporalType)
        {
            return SetParameter(position, value, temporalType);
        }

        private TypedQuery<X> SetParameter(int position, System.DateTime value,
                           TemporalType temporalType)
        {
            try
            {
                if (temporalType == TemporalType.DATE)
                {
                    m_hibernateQuery.SetDateTime(position, value);
                }
                else if (temporalType == TemporalType.TIME)
                {
                    m_hibernateQuery.SetTime(position, value);
                }
                else if (temporalType == TemporalType.TIMESTAMP)
                {
                    m_hibernateQuery.SetTimestamp(position, value);
                }

                Parameter<System.DateTime> p = new ParameterImpl<System.DateTime>(position, value.GetType());
                registerParameterBinding(p, value);
                return this;
            }
            catch (QueryParameterException e)
            {
                throw new ArgumentException(e.Message, e);
            }
            catch (HibernateException he)
            {
                throw m_entityManager.convert(he);
            }
        }

        private void registerParameterBinding<T>(Parameter<T> parameter, T value) 
	    {
            // not supported since CriteriaAPI is not supported
            /* 
		    if ( value != null && parameter.GetParameterType() != null ) 
		    {
    			if ( value is System.Collections.ICollection )
                {
    				ICollection collection = (ICollection) value;
				    // validate the elements...
				    foreach ( Object element  in collection ) {
                        if (!(parameter.GetParameterType().IsAssignableFrom(element.GetType())))
                        {
						    throw new ArgumentException(
								    "Parameter value [" + element + "] was not matching type [" +
										    parameter.GetParameterType().Name + "]"
						    );
					    }
				    }
			    }
			    else if ( value.GetType().IsArray) 
			    {
    				
				    Object[] array = (Object[]) value;
				    foreach ( Object element in array ) 
                    {
                        if (!(parameter.GetParameterType().IsAssignableFrom(element.GetType())))
                        {
						    throw new ArgumentException(
								    "Parameter value [" + element + "] was not matching type [" +
										    parameter.GetParameterType().Name + "]"
						    );
					    }
				    }
			    }
			    else 
                {
                    if (!(parameter.GetParameterType().IsAssignableFrom(value.GetType())))
                    {
					    throw new ArgumentException(
						    	"Parameter value [" + value + "] was not matching type [" +
									    parameter.GetParameterType().Name+ "]"
					    );
				    }
			    }
		    }
    */
		    if ( m_parameterBindings == null ) 
            {
                m_parameterBindings = new Dictionary<Parameter<Object>, Object>();
		    }
            Parameter<Object> p;
            if (!(typeof(T).FullName.Equals(typeof(Object).FullName)))
            {
                p = new ParameterImpl<Object>(parameter.GetName(), parameter.GetPosition(), value.GetType());
            }
            else 
            {
                p = (Parameter<Object>)parameter;
            }
            m_parameters.Add(p);
            m_parameterBindings[p] = value;
	    }

        /**
         * Get the parameter objects corresponding to the declared
         * parameters of the query.
         * Returns empty set if the query has no parameters.
         * This method is not required to be supported for native
         * queries.
         * @return set of the parameter objects
         * @throws InvalidOperationException if invoked on a native
         *         query when the implementation does not support 
         *         this use
         * @since Java Persistence 2.0
         */
        public HashSet<object> GetParameters()
        {
            throw new NotSupportedException();
        }

        /**
         * Get the parameter object corresponding to the declared
         * parameter of the given name.
         * This method is not required to be supported for native
         * queries.
         * @param name  parameter name
         * @return parameter object
         * @throws ArgumentException if the parameter of the
         *         specified name does not exist
         * @throws InvalidOperationException if invoked on a native
         *         query when the implementation does not support 
         *         this use
         * @since Java Persistence 2.0
         */
        public Parameter<T> GetParameter<T>(string name)
        {
            if ( name == null ) 
            {
			    throw new ArgumentException( "Name of parameter to locate cannot be null" );
		    }
            
            foreach (Parameter<Object> p in m_parameterBindings.Keys) 
            {
                if (name.Equals((p).GetName()))
                {
                    if ((typeof(T).FullName.Equals(typeof(Object).FullName)))
                    {
                        return (Parameter<T>)p;
                    }
                    return new ParameterImpl<T> (p.GetName(), p.GetParameterType());
                }
		    }            
		    throw new ArgumentException( "Unable to locate parameter named [" + name + "] of type " + typeof(T) );
        }

        /**
         * Get the parameter object corresponding to the declared
         * parameter of the given name and type.
         * This method is required to be supported for criteria queries
         * only.
         * @param name  parameter name
         * @param type  type
         * @return parameter object
         * @throws ArgumentException if the parameter of the
         *         specified name does not exist or is not assignable
         *         to the type
         * @throws InvalidOperationException if invoked on a native
         *         query or Java Persistence query language query when
         *         the implementation does not support this use
         * @since Java Persistence 2.0
         */
        public Parameter<T> GetParameter<T>(string name, System.Type type)
        {
            Parameter<T> param = GetParameter<T>(name);
            if (param.GetParameterType() != null && !param.GetParameterType().Equals(type))
            {
                throw new ArgumentException(
                        "Parameter type [" + param.GetParameterType().Name +
                                "] is not assignment compatible with requested type [" +
                                type.Name + "]"
                );
            }
            return param;
        }

        /**
         * Get the parameter object corresponding to the declared
         * positional parameter with the given position.
         * This method is not required to be supported for native
         * queries.
         * @param position  position
         * @return parameter object
         * @throws ArgumentException if the parameter with the
         *         specified position does not exist
         * @throws InvalidOperationException if invoked on a native
         *         query when the implementation does not support 
         *         this use
         * @since Java Persistence 2.0
         */
        public Parameter<T> GetParameter<T>(int position)
        {
            foreach (Parameter<Object> p in m_parameterBindings.Keys) 
            {
                if (p.GetPosition().HasValue && position == p.GetPosition().Value)
                {
                    if ((typeof(T).FullName.Equals(typeof(Object).FullName)))
                    {
                        return (Parameter<T>)p;
                    }
                    return new ParameterImpl<T>(p.GetPosition().Value, p.GetParameterType());
                }
		    }     
            
			throw new ArgumentException( "Unable to locate parameter with position [" + position + "]" );
             
        }

        /**
         * Get the parameter object corresponding to the declared
         * positional parameter with the given position and type.
         * This method is not required to be supported by the provider.
         * @param position  position
         * @param type  type
         * @return parameter object
         * @throws ArgumentException if the parameter with the
         *         specified position does not exist or is not assignable
         *         to the type
         * @throws InvalidOperationException if invoked on a native
         *         query or Java Persistence query language query when
         *         the implementation does not support this use
         * @since Java Persistence 2.0
         */
        public Parameter<T> GetParameter<T>(int position, System.Type type)
        {
            Parameter<T> param = GetParameter<T>(position);
            if (param.GetParameterType() != null && !param.GetParameterType().Equals(type))
            {
                // we were able to determine the expected type during analysis, so validate it here
                throw new ArgumentException(
                        "Parameter type [" + param.GetParameterType().Name +
                                "] is not assignment compatible with requested type [" +
                                type.Name + "]"
                );
            }
            return param;
        }

        /**
         * Return a boolean indicating whether a value has been bound 
         * to the parameter.
         * @param param parameter object
         * @return boolean indicating whether parameter has been bound
         * @since Java Persistence 2.0
         */
        public bool IsBound<T>(Parameter<T> param)
        {
            if (m_parameterBindings == null || param == null)
            {
                return false;
            }
            Parameter<T> p;
            try
            {
                if (param.GetName() != null)
                {
                    p = GetParameter<T>(param.GetName());
                }
                else
                {
                    p = GetParameter<T>(param.GetPosition().Value);
                }
            }
            catch (ArgumentException e)
            {
                return false;
            }

            return p != null;
        }

        /**
         * Return the value bound to the parameter.
         * @param param parameter object
         * @return parameter value
         * @throws ArgumentException if the parameter is not 
         *         a parameter of the query
         * @throws InvalidOperationException if the parameter has not been
         *         been bound
         * @since Java Persistence 2.0
         */
        public T GetParameterValue<T>(Parameter<T> param)
        {
            if (m_parameterBindings == null)
            {
                throw new InvalidOperationException("No parameters have been bound");
            }
            try
            {
                Parameter<Object> objParam;
                try
                {
                    if (param.GetName() != null)
                    {
                        objParam = GetParameter<Object>(param.GetName());
                    }
                    else
                    {
                        objParam = GetParameter<Object>(param.GetPosition().Value);
                    }
                }
                catch (ArgumentException e)
                {
                    throw new InvalidOperationException("Parameter has not been bound");
                }

                if (objParam != null)
                {
                    return (T)m_parameterBindings[objParam];
                }

                throw new InvalidOperationException("Parameter has not been bound");
            }
            catch (InvalidCastException e)
            {
                throw new InvalidOperationException("Encountered a parameter value type exception");
            }
        }

        /**
         * Return the value bound to the named parameter.
         * @param name  parameter name
         * @return parameter value
         * @throws InvalidOperationException if the parameter has not been
         *         been bound
         * @throws ArgumentException if the parameter of the
         *         specified name does not exist
         * @since Java Persistence 2.0
         */
        public object GetParameterValue(string name)
        { 
            return GetParameterValue( GetParameter<Object>( name ) );
        }

        /**
         * Return the value bound to the positional parameter.
         * @param position  position
         * @return parameter value
         * @throws InvalidOperationException if the parameter has not been
         *         been bound
         * @throws ArgumentException if the parameter with the
         *         specified position does not exist
         * @since Java Persistence 2.0
         */
        public object GetParameterValue(int position)
        {
            return GetParameterValue(GetParameter<Object>(position));
        }

        /**
         * Set the flush mode type to be used for the query execution.
         * The flush mode type applies to the query regardless of the
         * flush mode type in use for the entity manager.
         * @param flushMode  flush mode
         * @return the same query instance
         */
        TypedQuery<X> TypedQuery<X>.SetFlushMode(FlushModeType flushMode)
        {
            return SetFlushMode(flushMode);
        }

        Query Query.SetFlushMode(FlushModeType flushMode)
        {
            return SetFlushMode(flushMode);
        }

        private TypedQuery<X> SetFlushMode(FlushModeType flushMode)
        {
            if (flushMode == FlushModeType.AUTO)
            {
                m_hibernateQuery.SetFlushMode(FlushMode.Auto);
            }
            else if (flushMode == FlushModeType.COMMIT)
            {
                m_hibernateQuery.SetFlushMode(FlushMode.Commit);
            }
            else
            {
                throw new ArgumentException("Unexpected argument: " + flushMode);
            }
            m_flushMode = flushMode;
            return this;
        }

        

        /**
         * Get the flush mode in effect for the query execution. 
         * If a flush mode has not been set for the query object, 
         * returns the flush mode in effect for the entity manager.
         * @return flush mode
         * @since Java Persistence 2.0
         */
        public FlushModeType GetFlushMode()
        {
            if (m_flushMode.HasValue)
            {
                return m_flushMode.Value;
            }
            return m_entityManager.GetFlushMode();
        }

        /**
         * Set the lock mode type to be used for the query execution.
         * @param lockMode  lock mode
         * @return the same query instance
         * @throws InvalidOperationException if the query is found not to be 
         *         a Java Persistence query language SELECT query
         *         or a Criteria API query
         * @since Java Persistence 2.0
         */
        TypedQuery<X> TypedQuery<X>.SetLockMode(LockModeType lockMode)
        {
            return SetLockMode(lockMode);
        }

        Query Query.SetLockMode(LockModeType lockMode)
        {
            return SetLockMode(lockMode);
        }

        private TypedQuery<X> SetLockMode(LockModeType lockMode)
        {
            if (m_entityManager.isTransactionInProgress())
            {
                throw new TransactionRequiredException("no transaction is in progress");
            }
            LockMode lm = LockModeTypeHelper.getLockMode(lockMode);
            m_hibernateQuery.SetLockMode("this", lm);
            m_lockMode = lockMode;
            return this;
        }



        /**
         * Get the current lock mode for the query.
         * @return lock mode
         * @throws InvalidOperationException if the query is found not to be
         *         a Java Persistence query language SELECT query or
         *         a Criteria API query
         * @since Java Persistence 2.0
         */
        public LockModeType GetLockMode()
        {
            return m_lockMode;
        }

        /**
         * Return an object of the specified type to allow access to 
         * the provider-specific API.  If the provider's query 
         * implementation does not support the specified class, the 
         * <code>PersistenceException</code> is thrown.
         * @param cls  the class of the object to be returned.  This is
         *             normally either the underlying query 
         *             implementation class or an interface that it 
         *             implements.
         * @return an instance of the specified class
         * @throws PersistenceException if the provider does not support
         *         the call
         * @since Java Persistence 2.0
         */
        public T Unwrap<T>(System.Type cls)
        {
            if (cls.IsAssignableFrom(m_hibernateQuery.GetType()))
            {
                return (T)m_hibernateQuery;
            }
		    else 
		    {
                PersistenceException pe = new PersistenceException("Unsupported unwrap target type [" + cls.Name + "]");
				throw pe;
			    
		    }
        }
    }
}
