﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using NPersistence;
using NHibernate;
using NPersistence.NHibernate.util;

namespace NPersistence.NHibernate.transaction
{
    class EntityTransactionImpl : EntityTransaction
    {
        EntityManagerImpl m_entityManagerImpl;
        private ITransaction m_transaction = null;
        private bool m_rollbackOnly;

        public EntityTransactionImpl(EntityManagerImpl entityManagerImpl)
        {
            m_entityManagerImpl = entityManagerImpl;
        }

        private ISession getSession()
        {
            return m_entityManagerImpl.getSession();
        }

        public void Begin()
        {
            try
            {
                m_rollbackOnly = false;
                if (m_transaction != null && m_transaction.IsActive)
                {
                    throw new InvalidOperationException("Transaction already active");
                }
                //entityManager.adjustFlushMode();
                m_transaction = getSession().BeginTransaction();
            }
            catch (HibernateException he)
            {
                m_entityManagerImpl.throwPersistenceException(he);
            }
        }

        public void Commit()
        {
            if ( m_transaction == null || !m_transaction.IsActive ) 
            {
                throw new InvalidOperationException("Transaction not active");
		    }
		    if ( m_rollbackOnly ) 
            {
			    m_transaction.Rollback();
			    throw new RollbackException( "Transaction marked as rollbackOnly" );
		    }
		    try 
            {
			    m_transaction.Commit();
		    }
		    catch (Exception e) 
            {
			    Exception wrappedException;
			    if (e is StaleStateException) 
                {
				    wrappedException = m_entityManagerImpl.wrapStaleStateException( (StaleStateException) e );
			    }
			    else 
                {
				    wrappedException = e;
			    }
			    try 
                {
				    //as per the spec we should rollback if commit fails
                    m_transaction.Rollback();
			    }
			    catch (Exception re) {
				    //swallow
			    }
			    throw new RollbackException( "Error while commiting the transaction", wrappedException );
		    }
		    finally 
            {
			    m_rollbackOnly = false;
		    }
        }

        public void Rollback()
        {
            if (m_transaction == null || !m_transaction.IsActive)
            {
                throw new InvalidOperationException("Transaction not active");
            }
            try
            {
                m_transaction.Rollback();
            }
            catch (Exception e)
            {
                throw new PersistenceException("unexpected error when rollbacking", e);
            }
            finally
            {
                try
                {
                    if (m_entityManagerImpl != null)
                    {
                        ISession session = getSession();
                        if (session != null && session.IsOpen)
                        {
                            session.Clear();
                        }
                    }
                }
                catch (Exception t)
                {
                    //we don't really care here since it's only for safety purpose
                }
                m_rollbackOnly = false;
            }
        }

        /**
         * Mark the current resource transaction so that the only 
         * possible outcome of the transaction is for the transaction 
         * to be rolled back. 
         * @throws InvalidOperationException if <code>isActive()</code> is false
         */
        public void SetRollbackOnly()
        {
            if (!IsActive())
            {
                throw new InvalidOperationException("Transaction not active");
            }
            m_rollbackOnly = true;
        }

        /**
         * Determine whether the current resource transaction has been 
         * marked for rollback.
         * @return boolean indicating whether the transaction has been
         *         marked for rollback
         * @throws InvalidOperationException if <code>isActive()</code> is false
         */
        public bool GetRollbackOnly()
        {
            if (!IsActive())
            {
                throw new InvalidOperationException("Transaction not active");
            }
            return m_rollbackOnly;
        }

        /**
         * Indicate whether a resource transaction is in progress.
         * @return boolean indicating whether transaction is
         *         in progress
         * @throws PersistenceException if an unexpected error 
         *         condition is encountered
         */
        public bool IsActive()
        {
            try
            {
                return m_transaction != null && m_transaction.IsActive;
            }
            catch (Exception e)
            {
                throw new PersistenceException("unexpected error when checking transaction status", e);
            }
        }

    }
}
